#include <QEventLoop>
#include <QScriptEngine>
#include <QByteArray>
#include <QSettings>
#include <QStringList>
#include <QApplication>
#include <QUrl>
#include <QFile>
#include <QDebug>
#include <QMessageBox>
#include <QDesktopWidget>

#include "logger.h"
#include "ui/widgets/browser/browser.h"
#include "ui/widgets/document.h"
#include "ui/widgets/gif.h"
#include "ui/widgets/idle.h"
#include "ui/widgets/image.h"
#include "ui/widgets/pdf.h"
#include "ui/widgets/video.h"
#include "ui/widgets/webpage.h"
#include "ui/widgets/externalbrowser.h"
#include "playlistparser.h"
#include "downloader.h"
#include "dispatcher.h"

#include <QProcess>
#include "ui/exitpassworddialog.h"
#include "utils.h"

Dispatcher::Dispatcher(Viewer *viewer, const CommandLineParser *cmdparse)
    : QObject()
    , m_status(Dispatcher::StatusNone)
    , m_currentWidget(0)
    , m_prevWidget(0)
    , m_viewer(viewer)
    , m_plRefresh(false)
    , m_idling(false)
    , m_is_valid(false)
{
    settings = Settings::load(false, cmdparse);

//    if (settings->versionChanged())
//    {
//        QMessageBox::warning(
//            viewer,
//            tr("Software version changed"),
//            tr("You changed software version. Please reset your linutop settings."));
//        m_is_valid = false;
//    }
//    else
        m_is_valid = true;
        start_sec = QDateTime::currentDateTime();

    remote = new Remote(this);
    cleanser = new Cleanser(settings);
    QObject::connect(remote, SIGNAL(restart()),            this,     SLOT(startSlideshow()));
    QObject::connect(remote, SIGNAL(next()),               this,     SLOT(next()));
    QObject::connect(remote, SIGNAL(goTo(int,int)),        this,     SLOT(goTo(int,int)));
    QObject::connect(remote, SIGNAL(goTo(QString)),        this,     SLOT(goTo(QString)));
    QObject::connect(remote, SIGNAL(pause()),              m_viewer, SLOT(pause()));
    QObject::connect(remote, SIGNAL(screenshot()),         this,     SLOT(shootScreen()));
    QObject::connect(remote, SIGNAL(text(QString,int)),    m_viewer, SLOT(showText(QString,int)));
    QObject::connect(remote, SIGNAL(readUrl(QString,int)), m_viewer, SLOT(showUrl(QString,int)));

    m_mode = settings->get("main/mode").toString();

    QObject::connect(viewer, SIGNAL(restart()),           this, SLOT(startSlideshow()));
    QObject::connect(viewer, SIGNAL(end()),               this, SLOT(next()));
    QObject::connect(viewer, SIGNAL(mouseEvent()),        this, SLOT(onMouseEvent()));
    QObject::connect(viewer, SIGNAL(terminate()),         this, SLOT(finalize()));
    QObject::connect(viewer, SIGNAL(startInlineRemote()), this, SLOT(startInlineRemote()));
    QObject::connect(viewer, SIGNAL(readyToShoot()),      this, SLOT(shootScreen()));
#ifdef __arm__
    endIdle();
    connect(viewer, SIGNAL(terminate()),         this, SLOT(endIdle()));
#endif

    if(settings->get("main/fullscreen").toBool())
    {
        QRect rec = QApplication::desktop()->screenGeometry();
        int height = rec.height();
        int width = rec.width();
        m_viewer->setFixedSize(width,height);
    }
    // Disable screensaver when option is selected.
    if (settings->get("main/no_screensaver").toBool())
    {
        QProcess p;
//        QStringList args;
//        args << "xscreensaver-command";
//        QObject::connect(
//                    &p,   SIGNAL(finished(int, QProcess::ExitStatus)),
//                    this, SLOT(startNoScreensaver(int, QProcess::ExitStatus)));
//        p.start("which", args);
//        p.waitForFinished();

        //Disable xscreensaver deamon and suppress ctrl alt suppr shortcut
        p.start("pkill -9 xscreensaver");
        p.waitForFinished();
    }

    // Delete previous screenshot if needed.
    {
        if (QFile::exists("/tmp/screen.png"))
            QFile::remove("/tmp/screen.png");
    }
}

Dispatcher::~Dispatcher(void)
{
    // Kill the created server.
    if (!remotePid.isEmpty())
    {
        QProcess cmdKill;
        cmdKill.start("sudo", QStringList() << "kill" << "-9" << remotePid);
        cmdKill.waitForFinished();
        qDebug() << "Remote killed:" << remotePid;
    }
    else
    {
        qDebug() << "No remote to kill.";
    }
}

void Dispatcher::startNoScreensaver(
        int exitCode, QProcess::ExitStatus exitStatus)
{ if (exitCode == 0) disableScreensaver(); }

void Dispatcher::onScreensaverTimeout()
{
    QProcess p;
    QStringList args;
    args << "-deactivate";
    p.start("xscreensaver-command", args);
    p.waitForFinished();
}

void Dispatcher::disableScreensaver()
{
    connect(&m_ssTimer, SIGNAL(timeout()), this, SLOT(onScreensaverTimeout()));
    m_ssTimer.start(30000);
}


void Dispatcher::onPlTimeout()
{
    m_plRefresh = true;
    QString updateMode = settings->get(
                "slideshow/playlist_update_mode").toString();
    if (updateMode != "localPlaylist")
    {
        QDateTime lastEdit = getLastEdit();
        if(lastEdit<m_lastUpdate) m_plRefresh = false;
    }
}

bool Dispatcher::startup(void)
{
    connect(&m_plTimer, SIGNAL(timeout()), this, SLOT(onPlTimeout()));
    int timeout = settings->get("slideshow/refresh_after_minutes").toInt();
    m_plTimer.start(60000 * timeout);

    if (m_mode == "kiosk" && settings->get("main/start_with_session").toBool())
        return startBrowser();
    else
        startSlideshow();

    return true;
}

void Dispatcher::onMouseEvent()
{
    if (m_mode != "kiosk")
        return;

    m_currentWidget->stop();
    m_loopIdx = -1;
    m_loopIdxList.replace(m_listIdx, m_loopIdx);
    startBrowser();
}

void Dispatcher::onReconnectMouseEventTimeout()
{
    connect(m_viewer, SIGNAL(mouseEvent()), this, SLOT(onMouseEvent()));
}

void Dispatcher::reconnectMouseEvent()
{
    QTimer::singleShot(1000, this, SLOT(onReconnectMouseEventTimeout()));
}

bool Dispatcher::next(void)
{
    if (m_status == Dispatcher::StatusNone) return startup();

    else if (m_status == Dispatcher::StatusBrowser)
    {
        qDebug() << "dispatcher: End StatusBrowser";
        disconnect(m_viewer, SIGNAL(mouseEvent()), this, SLOT(onMouseEvent()));
        execHook("session-stop");
        afterBrowser();
    }

    else if (m_status == Dispatcher::StatusAfterBrowser)
    {
        qDebug() << "dispatcher: End StatusAfterBrowser";
        if (settings->get("main/after_web_session").toString() == "session")
        {
            Logger::logInfo("STARTING: web session");
            startBrowser();
        }
        else
        {
            reconnectMouseEvent();
            startSlideshow();
        }
    }
    else showLoopNext();

    return true;
}

void Dispatcher::shootScreen(void)
{
    if (!m_viewer->currentWidget()) return;

    QPixmap::grabWidget(m_viewer).save("/tmp/screen.png", "PNG");
    qDebug() << "Screenshot saved at: /tmp/screen.png";
}

bool Dispatcher::view_next()
{
    return m_viewer->next(dynamic_cast<QWidget *>(m_currentWidget));
//    if (m_prevWidget)
//    {
//        qDebug() << "DELETE CALLED";
//        dynamic_cast<QWidget *>(m_prevWidget)->deleteLater();
//        m_prevWidget = NULL;
//    }
}

void Dispatcher::afterBrowser()
{
    m_status = Dispatcher::StatusAfterBrowser;
    QString uri = settings->get("main/post_session_uri").toString();
    if (uri.isEmpty())
    {
        next();
        return;
    }
    int time = settings->get("main/post_session_uri_time").toInt();
    PlaylistItem item(uri, PlaylistItem::TypeWebpage, time);
    m_currentWidget = new Webpage(&item, m_viewer);
    view_next();
}

bool Dispatcher::startBrowser(QUrl url)
{
    if (settings->get("main/external_browser").toBool())
    {
        bool use_password = settings->get("main/admin_passwd").toBool();

        QScopedPointer<panel::Manager> panelManager(
                    panel::Manager::create(true, use_password));
        QScopedPointer<SystemShortcutManager> shortcutManager(
                    new SystemShortcutManager(true, use_password));

        QProcess process;
        process.start(settings->get("main/external_browser_cmd").toString() +
                      " " + settings->get("main/default_url").toString());
        m_status = Dispatcher::StatusBrowser;
        process.waitForFinished(-1);

        if (use_password)
        {
            ExitPasswordDialog d;
            QString password = d.run();
            if (password.isEmpty() || !Utils::validateAdminPassword(password))
                startBrowser();
        }

        return false;
    }
    else
    {
        execHook("session-start");
        Logger::logInfo("STARTING: web session");
        m_status = Dispatcher::StatusBrowser;
        m_currentWidget = new Browser;
        if (!url.isEmpty())
            static_cast<Browser*>(m_currentWidget)->setHome(url.toString());
        view_next();
    }

    return true;
}

void Dispatcher::execHook(QString hook)
{
    QString hookFile = settings->hookDirname() + QString("/%1").arg(hook);
    QProcess p;
    p.start(hookFile);
    p.waitForFinished();
}

QDateTime Dispatcher::getLastEdit()
{
    QUrl url = QUrl(settings->get("slideshow/pages").toString());
    qDebug() << "Loading pages from:" << url.toString().toStdString().c_str();

    QUrl urlReq;
    urlReq.setScheme(url.scheme());
    urlReq.setHost(url.host());
    urlReq.setPath("/webapp/db/lastEdit/playlist/"+((XmlPlaylistParser *)m_pl)->getName());

    qDebug() << "Request URL: " << urlReq.toString().toStdString().c_str();

    QNetworkRequest req(urlReq);
    QNetworkAccessManager nam;
    QNetworkReply * reply = nam.get(req);

    QEventLoop loop;
    connect(reply, SIGNAL(finished()), &loop, SLOT(quit()));
    loop.exec();

    qDebug() << "Reply finished";

    QScriptEngine engine;
    QString res = reply->readAll();
    QScriptValue v = engine.evaluate(res);
    res = "";
    Q_FOREACH (QString number, v.toVariant().toStringList())
        res += number+";";

    return QDateTime::fromString(res, "yyyy;MM;dd;hh;mm;ss;");
}

void Dispatcher::startSlideshow()
{
    cleanser->check();
    if (settings->get("remote/enabled").toBool()) startRemote();

    QUrl url;

    settings = Settings::load(true, settings->getCommandLineParser());

    QString updateMode = settings->get(
                "slideshow/playlist_update_mode").toString();

    if (updateMode != "localPlaylist")
        url = QUrl(settings->get("slideshow/pages").toString());
    else
        url = QUrl(QString("file://%1").arg(settings->playlistFilename()));

    bool ok = false;
    int default_time = settings->get("slideshow/default_time").toInt(&ok);
    m_pl = PlaylistParser::get((ok ? default_time : 5), url, m_viewer);

    if (updateMode != "localPlaylist")
        m_lastUpdate = getLastEdit();

    //We fill the list which contains the index of current item in each loops
    if(m_loopIdxList.count() != m_pl->getLoops().last()->id().toInt() || updateMode == "localPlaylist")
    {
        if(m_loopIdxList.isEmpty())
        {
            qDebug() << "NB LOOPS : " << m_pl->getLoops().count();
            for(int i=0; i < m_pl->getLoops().count(); i++)
                m_loopIdxList.append(-1);
        }
    }

    m_loop = m_pl->selectLoop(start_sec);
    m_listIdx = m_loop->id().toInt();
    m_loopIdx = m_loopIdxList.at(m_listIdx);
    m_loopIdxList.replace(m_listIdx, m_loopIdx);

    Q_FOREACH (PlaylistLoop* loop, m_pl->getLoops())
    {
        if (loop->downloadable())
        {
            Downloader *d = new Downloader(loop);
            QEventLoop el;
            connect(d,SIGNAL(downloadDone()),&el,SLOT(quit()));
            d->start();
            if(!d->isDone())el.exec();
            qDebug()<<"Loop downloaded";
        }
        qDebug()<<"Loop downloaded";
    }
    showLoopNext();
}

void Dispatcher::goTo(int index, int loop)
{
    m_loop = m_pl->getLoops().at(loop);
    m_listIdx = m_loop->id().toInt();
    m_loopIdx = index - 1;
    m_loopIdxList.replace(m_listIdx, m_loopIdx);
    showLoopNext();
}

void Dispatcher::goTo(QString str)
{
    // We suspend the actual playlist and create a RAM one to display directly an item.
    QString content = QString("#Linutop playlist\nfile:/" + str + " 1\n");
    m_loop = new PlaylistLoop(1, content, QString(""));
    m_loopIdx = -1;
    m_loopIdxList.replace(m_listIdx, m_loopIdx);
    this->showLoopNext();
    m_viewer->pause();
}

void Dispatcher::showLoopNext()
{
    QString updateMode = settings->get(
                "slideshow/playlist_update_mode").toString();
    m_status = Dispatcher::StatusPlaylist;
    m_loop = m_pl->selectLoop(start_sec);
    if(updateMode != "localPlaylist")
        m_listIdx = m_loop->id().toInt();
    else
        m_listIdx = 0;

    m_loopIdx = m_loopIdxList.at(m_listIdx);

    ++m_loopIdx;
    m_loopIdxList.replace(m_listIdx, m_loopIdx);

    if (m_loopIdx >= m_loop->count())
    {
        /* We need to download the playlist to check what's changed */
        if (m_plRefresh)
        {
            m_plRefresh = false;
            startSlideshow();
            return;
        }
        m_pl->refresh();
        m_loop = m_pl->selectLoop(start_sec);
        m_listIdx = m_loop->id().toInt();
        if (m_loop->count() == 0)
        {
            m_loop->items().append(
                        new PlaylistItem(
                            "file://:/image/nothing_to_display.png",
                            PlaylistItem::TypeUndefined, 5, false));
        }
        m_loopIdx = -1;

        m_loopIdxList.replace(m_listIdx, m_loopIdx);
        if (m_loop->downloadable())
        {
            Downloader *d = new Downloader(m_loop);
            connect(d, SIGNAL(downloadDone()), this, SLOT(showLoopNext()));
            d->start();
            return;
        }
        else
            m_loopIdx = 0; /* no download, no event, so continue */
        m_loopIdxList.replace(m_listIdx, m_loopIdx);
    }

    PlaylistItem *item;
    if (m_loop->count() > m_loopIdx)
    {
        item = m_loop->at(m_loopIdx);
        start_sec = QDateTime::currentDateTime();

        if(item->uri().contains("Screen OFF")) //Mise en place idle sur la boucle locale
        {
            QDateTime start_idle = m_loop->getFromIdle(item->uri());
            QDateTime stop = m_loop->getToIdle(item->uri());
            QDateTime stop_idle = stop;

            if(stop_idle.time() < start_idle.time())
                stop_idle = stop.addDays(1);

            if (start_idle > start_sec || stop_idle < start_sec)
            {
                m_loopIdx++;
                m_loopIdxList.replace(m_listIdx, m_loopIdx);
                item = m_loop->at(m_loopIdx);
            }
            else
            {
                int new_time;
                int maxTime = 24*3600;
                int to_time = stop_idle.time().hour()*3600 + stop_idle.time().minute()*60;
                int now_time = start_sec.time().hour()*3600 + start_sec.time().minute()*60
                                +start_sec.time().second();

                if(start_idle.date().day() != stop_idle.date().day())
                    new_time = (maxTime - now_time) + to_time;
                else
                    new_time = to_time - now_time;

                qDebug() << new_time;
                item->setTime(new_time);
            }
        }
    }

    else
        return;

    qDebug() << item->uri() << item->type() << item->href();

    if (item->href())
    {
        showLoopNext();
        return;
    }

    m_prevWidget = m_currentWidget;

    switch (item->type())
    {
        case PlaylistItem::TypeIdle:
            m_currentWidget = new Idle(item, m_viewer);
            break;
        case PlaylistItem::TypeImage:
            m_currentWidget = new Image(item, m_viewer);
            break;
        case PlaylistItem::TypeGif:
            m_currentWidget = new Gif(item, m_viewer);
            break;
        case PlaylistItem::TypeVideo:
            m_currentWidget = new Video(item, m_viewer);
            break;
        case PlaylistItem::TypePdf:
            m_currentWidget = new PDF(item, m_viewer);
            break;
        case PlaylistItem::TypeWebpage:
            if (settings->get("slideshow/external_browser").toBool())
                m_currentWidget = new ExternalBrowser(item, m_viewer);
            else
                m_currentWidget = new Webpage(item, m_viewer);
            break;
        case PlaylistItem::TypeDocument:
            m_currentWidget = new Document(item, m_viewer);
            break;
        default:
            next();
            return;
    }

    if ((dynamic_cast<Idle *>(m_prevWidget) != NULL)
        && (item->type() != PlaylistItem::TypeIdle))
    {
        qDebug() << "Stopping idle";
        QProcess p;
#ifdef __arm__
        endIdle();
#else
        p.start("xset", QStringList() << "dpms" << "force" << "on");
        p.waitForFinished();
#endif
    }

    if (!view_next())
    {
        --m_loopIdx;
        showLoopNext();
    }
}

void Dispatcher::startRemote()
{
    // Kill remaining surviving servers.
    {
        QProcess cmdKill;
        cmdKill.start("sudo", QStringList() << "pkill" << "-9" << "linutop-remote");
        cmdKill.waitForFinished();
    }

    // Start the new server.
    {
        QStringList args;
#ifdef DEBUG
        args << LINUTOP_REMOTE_CMD << "--debug";
#else
        args << "linutop-remote";
#endif // DEBUG
        args << getenv("HOME");
        args << getenv("USER");
        qDebug() << "Start remote: sudo"
                 << args.join(" ").toStdString().c_str();
        QProcess server;
        server.setWorkingDirectory("/");
        server.start("sudo", args);
        server.waitForFinished();
        QByteArray output = server.readAllStandardOutput();
        remotePid = QString(output.data());

        qDebug() << "Remote started, pid:"
                 << remotePid.toStdString().c_str();
    }
}

void Dispatcher::finalize()
{
    if (m_currentWidget)
        dynamic_cast<QWidget*>(m_currentWidget)->close();
    qApp->quit();
}

void Dispatcher::endIdle()
{
    QProcess p;
    p.start("sudo", QStringList() << "/opt/vc/bin/tvservice" << "--status");
    p.waitForFinished();
    QByteArray output = p.readAllStandardOutput();

    if (output.contains("[TV is off]"))
    {
        p.start("sudo", QStringList() << "/opt/vc/bin/tvservice" << "-p");
        p.waitForFinished();
        p.start("sudo", QStringList() << "chvt" << "1");
        p.waitForFinished();
        p.start("sudo", QStringList() << "chvt" << "7");
        p.waitForFinished();
    }
}

void Dispatcher::startInlineRemote()
{
    m_viewer->pause();
    QUrl remote_url(settings->get("remote/local_url").toString() +
         "?k=" + settings->get("remote/key").toString());
    qDebug() << remote_url;
    startBrowser(remote_url);
    dynamic_cast<Browser*>(m_currentWidget)->hideToolbar();
}
