#include <cassert>

#include <QWebSettings>
#include <QSqlDatabase>
#include <QSettings>
#include <QShortcut>
#include <QEventLoop>
#include <QDesktopWidget>
#include <QWebFrame>
#include <QWebElement>
#include <QCursor>

#include <QDebug>

#include "ui/exitpassworddialog.h"
#include "ui/widgets/browser/browser.h"
#include "ui/widgets/webpage.h"
#include "viewer.h"
#include "utils.h"

PlaylistEventFilter::PlaylistEventFilter()
    : QObject()
    , m_paused(true)
{
    settings = Settings::load();
}

bool PlaylistEventFilter::eventFilter(QObject *obj, QEvent *event)
{
    if (m_paused) {
        if (event->type() == QEvent::MouseButtonPress) {
            return true;
        }
        return false;
    }

    switch (event->type())
    {
        case QEvent::MouseMove:
            if (settings->get("remote/enabled").toBool())
                return false;
        case QEvent::MouseButtonPress:
            if (settings->get("remote/enabled").toBool())
                return true;
        case QEvent::KeyPress:
            qobject_cast<QWidget *>(obj)->setCursor(Qt::BlankCursor);
            /* hum, not nice... */
            qobject_cast<Viewer *>(obj->parent())->emitMouseEvent();
            return true;
        default:
            return false;
    }
}

bool BrowserEventFilter::eventFilter(QObject *obj, QEvent *event)
{
    QMouseEvent *me = dynamic_cast<QMouseEvent *>(event);
    QInputEvent *ie = dynamic_cast<QInputEvent *>(event);
    if (me || ie)
        qobject_cast<Viewer *>(obj->parent())->resetTimer();

    return false;
}

Viewer::Viewer()
    : QStackedWidget()
    , m_playlistEventFilter(NULL)
    , m_browserEventFilter(NULL)
{
    settings = Settings::load();
    prevWidget = 0;

    connect(&m_timer, SIGNAL(timeout()), this, SLOT(onTimerTimeout()));

    m_playlistEventFilter = new PlaylistEventFilter;
    m_browserEventFilter = new BrowserEventFilter;
    m_unlockTimer.setSingleShot(true);
    connect(&m_unlockTimer, SIGNAL(timeout()),
            this, SLOT(unlockPlaylistEventFilter()));

    loadWindowSettings();
    setShortcuts();

    // Security setup.
    {
        bool fullscreen = settings->get("main/fullscreen").toBool();
        bool use_password = settings->get("main/admin_passwd").toBool();

        QScopedPointer<panel::Manager> tmp1(
            panel::Manager::create(fullscreen, use_password));
        m_panel_manager.swap(tmp1);

        QScopedPointer<SystemShortcutManager> tmp2(
            new SystemShortcutManager(fullscreen, use_password));
        m_system_shortcut_manager.swap(tmp2);

        if (fullscreen) showFullScreen();
        else showNormal();
    }

    // Hide mouse.
    if (settings->get("slideshow/hide_pointer").toBool())
    {
        qDebug() << "Hiding mouse pointer.";
        QScopedPointer<QProcess> hide(new QProcess);
        hide->start("unclutter", QStringList() << "-idle" << "1");
        m_hideMouse.swap(hide);
    }

    // Usb key detection setup.
    {
        if (settings->get("slideshow/usb_auto").toBool() &&
            (settings->get("main/mode").toString() == "display"))
        {
            QString remoteKey = settings->get("remote/key").toString();
            QScopedPointer<UsbAuto> usbauto(new UsbAuto(remoteKey));
            m_usbauto.swap(usbauto);
            QObject::connect(m_usbauto.data(), SIGNAL(finished()),
                             this, SLOT(requestRestart()));
        }
    }

    // Mute setting.
    {
        QStringList outputList;
        #ifndef __arm__
        outputList << "Master" << "Headphone" << "Speaker" << "PCM";
        #else
        outputList << "PCM";
        #endif

        if (settings->get("slideshow/mute").toBool())
        {
            Q_FOREACH(QString output, outputList)
            {
                qDebug() << "Disabling sound on channel:" << output;
                QStringList args;
                args << "set" << output << "mute";
                QProcess mute;
                mute.start("amixer", args);
                mute.waitForFinished();
            }
        }
        else
        {
            Q_FOREACH(QString output, outputList)
            {
                qDebug() << "Enabling sound on channel:" << output;
                QStringList args;
                args << "set" << output << "unmute";
                QProcess mute;
                mute.start("amixer", args);
                mute.waitForFinished();
            }
        }

    }

    // Web browser setup.
    {
        QWebSettings *websettings = QWebSettings::globalSettings();
        websettings->setAttribute(QWebSettings::JavascriptCanOpenWindows, true);
        websettings->setAttribute(QWebSettings::PluginsEnabled, true);
        websettings->setAttribute(QWebSettings::JavaEnabled, true);



        QString fontStr = settings->get("webpage/default_font").toString();
        if (!fontStr.isEmpty())
        {
            QFont font; font.fromString(fontStr);
            websettings->setFontFamily(QWebSettings::StandardFont, font.family());
            websettings->setFontSize(QWebSettings::DefaultFontSize, font.pointSize());
        }
    }

    // Cookie database setup ?
    {
        QSqlDatabase db = QSqlDatabase::addDatabase("QSQLITE");
        db.setDatabaseName(settings->cookiesFilename());
        db.open();
    }
}

Viewer::~Viewer(void) {}

void Viewer::emitMouseEvent()
{
    Q_EMIT mouseEvent();
}

void Viewer::mousePressEvent(QMouseEvent *event) {
    if (event->button() == Qt::RightButton) {
        rightclick();
    }
}

void Viewer::loadWindowSettings()
{
    QSettings appsettings;
    appsettings.beginGroup("Viewer");
    resize(appsettings.value("size", QSize(800, 600)).toSize());
    move(appsettings.value("pos", QPoint(200, 200)).toPoint());
    appsettings.endGroup();
}

void Viewer::saveWindowSettings()
{
    QSettings appsettings;
    appsettings.beginGroup("Viewer");
    appsettings.setValue("size", size());
    appsettings.setValue("pos", pos());
    appsettings.endGroup();
}

void Viewer::setShortcuts()
{
    QShortcut *shortcut;

    shortcut = new QShortcut(QKeySequence(Qt::CTRL + Qt::Key_Q), this);
    shortcut->setContext(Qt::ApplicationShortcut);
    connect(shortcut, SIGNAL(activated()), this, SLOT(bye()));

    if (!settings->get("main/admin_passwd").toBool())
    {
        shortcut = new QShortcut(QKeySequence(Qt::Key_Escape), this);
        shortcut->setContext(Qt::ApplicationShortcut);
        connect(shortcut, SIGNAL(activated()), this, SLOT(bye()));
    }
}

void Viewer::bye()
{
    if (settings->get("main/admin_passwd").toBool())
    {
        ExitPasswordDialog d;
        QString password = d.run();
        if (password.isEmpty() || ! Utils::validateAdminPassword(password))
            return;
    }

    if (prevWidget)
    {
        Runnable *r = dynamic_cast<Runnable *>(prevWidget);
        if (r)
            r->stop();
    }
    saveWindowSettings();
    Q_EMIT terminate();
}

void Viewer::onTimerTimeout()
{
    m_timer.stop();
    Q_EMIT end();
}

void Viewer::stop() { m_timer.stop(); }

void Viewer::setTimer()
{
    m_timer.setInterval(settings->get("main/stop_session_after_seconds").toInt() * 1000);
    m_timer.start();
}

void Viewer::displayWebpage(Webpage *page)
{
    // Display webapage.
    {
        addWidget(page);
        QObject::connect(page, SIGNAL(rightclick()), this, SLOT(rightclick()));
        QObject::connect(page, SIGNAL(end()), this, SLOT(onWidgetTimeout()));
        page->show();
        setCurrentWidget(page);
    }

    // Stop running widget.
    Runnable *r = dynamic_cast<Runnable*>(prevWidget);
    if (r && r->isActive()) this->pause();

    // Wait page loaded.
    // TODO: fix next lines.
//    if (!page->isLoaded())
//    {
//        QEventLoop loop;
//        QObject::connect(page, SIGNAL(loadFinished(bool)), &loop, SLOT(quit()));
//        loop.exec();
//    }
    {
        QTimer ti;
        QEventLoop loop;
        QObject::connect(&ti, SIGNAL(timeout()), &loop, SLOT(quit()));

        ti.setSingleShot(true);
        ti.start(1000);

        loop.exec();
    }

    // Take screenshot.
    Q_EMIT readyToShoot();
}

void Viewer::resetTimer()
{
    if (this->m_timer.isActive()) {
        m_timer.start();
    }
}

void Viewer::closeEvent(QCloseEvent *event)
{
    bye();
    event->ignore();
}

bool Viewer::next(QWidget *widget)
{
    if (!widget) return false;

    addWidget(widget);
    widget->setMouseTracking(true);

    Browser *b = qobject_cast<Browser *>(widget);
    if (b)
    {
        if (settings->get("main/stop_session_after_seconds").toInt())
        {
            widget->installEventFilter(m_browserEventFilter);

            // Special case if the current page is the remote.
            if (b->getHome().toString() !=
                    settings->get("remote/local_url").toString() +
                    "?k=" + settings->get("remote/key").toString())
            {
                connect(widget, SIGNAL(mouseMoved()), this, SLOT(resetTimer()));
                setTimer();
            }
        }
    }
    else
    {
        widget->setCursor(Qt::BlankCursor);
        QCursor curs = widget->cursor();
        if(settings->get("slideshow/external_player").toBool() ||
                settings->get("slideshow/external_browser").toBool())
            curs.setPos(4000,4000);
        if (m_playlistEventFilter &&
            settings->get("slideshow/stop_on_event").toBool())
            widget->installEventFilter(m_playlistEventFilter);
    }

    QObject::connect(widget, SIGNAL(rightclick()), this, SLOT(rightclick()));
    QObject::connect(widget, SIGNAL(end()), this, SLOT(onWidgetTimeout()));

    /* VLC send a MouseMove event that might stop the slideshow, so temporarily
     * ignore the events
     */
    if (m_playlistEventFilter)
        m_playlistEventFilter->setPaused(true);

    m_unlockTimer.start(1000);
    setCurrentWidget(widget);
    QPalette plt;
    plt.setColor(QPalette::Window, Qt::black);
    widget->setPalette(plt);
    widget->setAutoFillBackground(true);
    widget->show();

    bool started = dynamic_cast<Runnable *>(widget)->start();
    if (!started) return false;

    if (prevWidget)
    {
        removeWidget(prevWidget);
        Runnable* run = dynamic_cast<Runnable*>(prevWidget);
        if (run) run->stop();
        qDebug() << "DELETE CALLED";
        prevWidget->deleteLater();
        prevWidget = NULL;
    }

    prevWidget = widget;
    if (widget->inherits(Webpage::staticMetaObject.className()))
    {
        QEventLoop loop;
        Webpage* page = qobject_cast<Webpage *>(widget);
        connect(page, SIGNAL(loadFinished(bool)), &loop, SLOT(quit()));
        loop.exec();
    }

    return true;
    //Q_EMIT readyToShoot();
}

void Viewer::pause()
{
    Runnable *r = dynamic_cast<Runnable *>(prevWidget);
    if (r) r->pause();
}

void Viewer::showText(const QString &text, int duration)
{
    // Create webpage to display text.
    Webpage* page;
    {
        qDebug() << "Displaying webpage:"
                 << QUrl::fromLocalFile(
                        "/usr/local/share/linutop-webkiosk/text.html")
                    .toString();
        qDebug() << "For" << duration << "seconds.";

        PlaylistItem* item = new PlaylistItem(
                    "file:///usr/share/linutop-webkiosk/text.html",
                    PlaylistItem::TypeWebpage,
                    duration);

        page = new Webpage(item, this);
        page->start();

        QEventLoop loop;
        QObject::connect(page, SIGNAL(loadFinished(bool)), &loop, SLOT(quit()));
        loop.exec();

        page->page()->mainFrame()->findFirstElement("#text").setPlainText(text);
    }

    displayWebpage(page);
}

void Viewer::showUrl(QString url, int duration)
{
    // Create webpage to display text.
    Webpage* page;
    {
        qDebug() << "Displaying webpage:" << QUrl(url).toString();
        qDebug() << "For" << duration << "seconds.";

        PlaylistItem* item = new PlaylistItem(
                    QUrl(url).toString(),
                    PlaylistItem::TypeWebpage,
                    duration);

        page = new Webpage(item, this);
        page->start();
    }

    displayWebpage(page);
}

void Viewer::unlockPlaylistEventFilter()
{
    m_playlistEventFilter->setPaused(false);
}

void Viewer::onWidgetTimeout()
{
    m_timer.stop();
    Q_EMIT end();
}

void Viewer::rightclick() {
    if (settings->get("remote/enabled").toBool())
        Q_EMIT startInlineRemote();
}

void Viewer::requestRestart()
{ Q_EMIT restart(); }
